@@ -30,6 +30,7 @@ from .auth import AuthenticatedClient, AuthMiddlewareFactory, ProxyClient
 from .logging import logger
 
 ACCEPT_HEADERS = {"Accept": "application/json"}
+ADLS_HEADERS = {"X-Ms-Blob-Type": "BlockBlob"}
 
 ELEMENT_TYPE_METADATA = "__element_type_metadata"
 ELEMENT_VALUE_METADATA = "__element_value_metadata"
@@ -44,6 +45,10 @@ class DominoError(Exception):
     """Base exception for known errors."""
 
 
+class UnauthenticatedError(DominoError):
+    """To handle exponential backoff."""
+
+
 def _unpack_flight_error(error: str) -> str:
     """Unpack a flight error message by remove extra information."""
     try:
@@ -130,6 +135,66 @@ class Config:
         return res
 
 
+@attr.s(auto_attribs=True)
+class ADLSConfig(Config):
+    """ADLS datasource configuration."""
+
+    container: Optional[str] = _config(elem=ConfigElem.BUCKET)
+
+    access_key: Optional[str] = _cred(elem=CredElem.PASSWORD)
+
+
+@attr.s(auto_attribs=True)
+class GCSConfig(Config):
+    """GCS datasource configuration."""
+
+    bucket: Optional[str] = _config(elem=ConfigElem.BUCKET)
+
+    private_key_json: Optional[str] = _cred(elem=CredElem.PASSWORD)
+
+
+@attr.s(auto_attribs=True)
+class GenericS3Config(Config):
+    """Generic S3 datasource configuration."""
+
+    bucket: Optional[str] = _config(elem=ConfigElem.BUCKET)
+    host: Optional[str] = _config(elem=ConfigElem.HOST)
+    region: Optional[str] = _config(elem=ConfigElem.REGION)
+
+    aws_access_key_id: Optional[str] = _cred(elem=CredElem.USERNAME)
+    aws_secret_access_key: Optional[str] = _cred(elem=CredElem.PASSWORD)
+
+
+@attr.s(auto_attribs=True)
+class MySQLConfig(Config):
+    """MySQL datasource configuration."""
+
+    database: Optional[str] = _config(elem=ConfigElem.DATABASE)
+
+    password: Optional[str] = _cred(elem=CredElem.PASSWORD)
+    username: Optional[str] = _cred(elem=CredElem.USERNAME)
+
+
+@attr.s(auto_attribs=True)
+class OracleConfig(Config):
+    """Oracle datasource configuration."""
+
+    database: Optional[str] = _config(elem=ConfigElem.DATABASE)
+
+    password: Optional[str] = _cred(elem=CredElem.PASSWORD)
+    username: Optional[str] = _cred(elem=CredElem.USERNAME)
+
+
+@attr.s(auto_attribs=True)
+class PostgreSQLConfig(Config):
+    """PostgreSQL datasource configuration."""
+
+    database: Optional[str] = _config(elem=ConfigElem.DATABASE)
+
+    password: Optional[str] = _cred(elem=CredElem.PASSWORD)
+    username: Optional[str] = _cred(elem=CredElem.USERNAME)
+
+
 @attr.s(auto_attribs=True)
 class RedshiftConfig(Config):
     """Redshift datasource configuration."""
@@ -155,7 +220,7 @@ class SnowflakeConfig(Config):
 
 @attr.s(auto_attribs=True)
 class S3Config(Config):
-    """S3 datasource configurationn."""
+    """S3 datasource configuration."""
 
     bucket: Optional[str] = _config(elem=ConfigElem.BUCKET)
     region: Optional[str] = _config(elem=ConfigElem.REGION)
@@ -164,7 +229,29 @@ class S3Config(Config):
     aws_secret_access_key: Optional[str] = _cred(elem=CredElem.PASSWORD)
 
 
-DatasourceConfig = Union[Config, RedshiftConfig, SnowflakeConfig, S3Config]
+@attr.s(auto_attribs=True)
+class SQLServerConfig(Config):
+    """SQL Server datasource configuration."""
+
+    database: Optional[str] = _config(elem=ConfigElem.DATABASE)
+
+    password: Optional[str] = _cred(elem=CredElem.PASSWORD)
+    username: Optional[str] = _cred(elem=CredElem.USERNAME)
+
+
+DatasourceConfig = Union[
+    ADLSConfig,
+    Config,
+    GCSConfig,
+    GenericS3Config,
+    MySQLConfig,
+    OracleConfig,
+    PostgreSQLConfig,
+    RedshiftConfig,
+    S3Config,
+    SQLServerConfig,
+    SnowflakeConfig,
+]
 
 
 @attr.s
@@ -262,7 +349,7 @@ class _Object:
             content: bytes content
         """
         signed_url = self.datasource.get_key_url(self.key, True)
-        res = httpx.put(signed_url, content=content)
+        res = httpx.put(signed_url, content=content, headers=self.datasource.headers())
         res.raise_for_status()
 
         self.datasource.client._log_metric(  # pylint: disable=protected-access
@@ -279,7 +366,7 @@ class _Object:
         """
         signed_url = self.datasource.get_key_url(self.key, True)
         with open(filename, "rb") as file:
-            res = httpx.put(signed_url, content=file)
+            res = httpx.put(signed_url, content=file, headers=self.datasource.headers())
         res.raise_for_status()
 
         content_size = os.path.getsize(filename)
@@ -296,7 +383,7 @@ class _Object:
             fileobj: bytes-like object or an iterable producing bytes.
         """
         signed_url = self.datasource.get_key_url(self.key, True)
-        res = httpx.put(signed_url, content=fileobj)
+        res = httpx.put(signed_url, content=fileobj, headers=self.datasource.headers())
         res.raise_for_status()
 
 
@@ -333,7 +420,7 @@ class Datasource:
         """Store configuration override for future query calls.
 
         Args:
-            config: One of S3Config, RedshiftConfig or SnowflakeConfig
+            config: specific datasource config class
         """
         self._config_override = config
 
@@ -367,6 +454,12 @@ class QueryDatasource(Datasource):
 class ObjectStoreDatasource(Datasource):
     """Represents a object store type datasource."""
 
+    def headers(self) -> Dict[str, str]:
+        """Return headers for http calls to blob signed URL."""
+        if self.datasource_type == DatasourceDtoDataSourceType.ADLSCONFIG.value:
+            return ADLS_HEADERS
+        return {}
+
     def Object(self, key: str) -> _Object:  # pylint: disable=invalid-name
         """Return an object with given key and datasource client."""
         return _Object(datasource=self, key=key)
@@ -405,12 +498,15 @@ class ObjectStoreDatasource(Datasource):
         Returns:
             Signed URL for given key
         """
-        return self.client.get_key_url(
-            self.identifier,
-            object_key,
-            is_read_write,
-            config=self._config_override.config(),
-            credential=self._config_override.credential(),
+        return cast(
+            str,
+            self.client.get_key_url(
+                self.identifier,
+                object_key,
+                is_read_write,
+                config=self._config_override.config(),
+                credential=self._config_override.credential(),
+            ),
         )
 
     def get(self, object_key: str) -> bytes:
@@ -475,9 +571,16 @@ class ObjectStoreDatasource(Datasource):
 
 
 DATASOURCES = {
-    DatasourceDtoDataSourceType.SNOWFLAKECONFIG: QueryDatasource,
+    DatasourceDtoDataSourceType.ADLSCONFIG: ObjectStoreDatasource,
+    DatasourceDtoDataSourceType.GCSCONFIG: ObjectStoreDatasource,
+    DatasourceDtoDataSourceType.GENERICS3CONFIG: ObjectStoreDatasource,
+    DatasourceDtoDataSourceType.MYSQLCONFIG: QueryDatasource,
+    DatasourceDtoDataSourceType.ORACLECONFIG: QueryDatasource,
+    DatasourceDtoDataSourceType.POSTGRESQLCONFIG: QueryDatasource,
     DatasourceDtoDataSourceType.REDSHIFTCONFIG: QueryDatasource,
     DatasourceDtoDataSourceType.S3CONFIG: ObjectStoreDatasource,
+    DatasourceDtoDataSourceType.SNOWFLAKECONFIG: QueryDatasource,
+    DatasourceDtoDataSourceType.SQLSERVERCONFIG: QueryDatasource,
 }
 
 
@@ -581,6 +684,7 @@ class DataSourceClient:
         logger.exception(message)
         raise Exception(message)
 
+    @backoff.on_exception(backoff.expo, UnauthenticatedError, max_time=60)
     def list_keys(
         self,
         datasource_id: str,
@@ -601,6 +705,7 @@ class DataSourceClient:
 
         Raises:
             Exception: if the response from the Proxy is not 200
+            UnauthenticatedError: if the request has invalid authentication
         """
         logger.info("list_keys", datasource_id=datasource_id, prefix=prefix)
 
@@ -616,11 +721,14 @@ class DataSourceClient:
 
         if response.status_code == 200:
             return cast(List[str], response.parsed)
+        if response.status_code == 401:
+            raise UnauthenticatedError
 
         error = cast(ProxyErrorResponse, response.parsed)
         logger.exception(error)
         raise Exception(f"{error.error_type}: {error.raw_error}")
 
+    @backoff.on_exception(backoff.expo, UnauthenticatedError, max_time=60)
     def get_key_url(  # pylint: disable=too-many-arguments
         self,
         datasource_id: str,
@@ -643,6 +751,7 @@ class DataSourceClient:
 
         Raises:
             Exception: if the response from the Proxy is not 200
+            UnauthenticatedError: if the request has invalid authentication
         """
         logger.info("get_key_url", datasource_id=datasource_id, object_key=object_key)
 
@@ -659,6 +768,8 @@ class DataSourceClient:
 
         if response.status_code == 200:
             return cast(str, response.parsed)
+        if response.status_code == 401:
+            raise UnauthenticatedError
 
         error = cast(ProxyErrorResponse, response.parsed)
         logger.exception(error)
@@ -678,7 +789,12 @@ class DataSourceClient:
             is_read_write: whether used for read or write
         """
         mode = LogMetricM.WRITE if is_read_write else LogMetricM.READ
-        type_map = {DatasourceDtoDataSourceType.S3CONFIG.value: LogMetricT.S3CONFIG}
+        type_map = {
+            DatasourceDtoDataSourceType.ADLSCONFIG.value: LogMetricT.ADLSCONFIG,
+            DatasourceDtoDataSourceType.GCSCONFIG.value: LogMetricT.GCSCONFIG,
+            DatasourceDtoDataSourceType.GENERICS3CONFIG.value: LogMetricT.GENERICS3CONFIG,  # noqa
+            DatasourceDtoDataSourceType.S3CONFIG.value: LogMetricT.S3CONFIG,
+        }
         type_ = type_map.get(datasource_type)
         if not type_:
             return
@@ -731,10 +847,6 @@ class DataSourceClient:
             raise DominoError(_unpack_flight_error(str(exc))) from None
         return Result(self, reader, query)
 
-    @backoff.on_exception(
-        backoff.expo,
-        flight.FlightUnauthenticatedError,
-        max_time=60,
-    )
+    @backoff.on_exception(backoff.expo, flight.FlightUnauthenticatedError, max_time=60)
     def _do_get(self, ticket: str) -> flight.FlightStreamReader:
         return self.proxy.do_get(flight.Ticket(ticket))
